@@ -3,11 +3,11 @@
  * nodeSubplan.c
  *	  routines to support subselects
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/executor/nodeSubplan.c,v 1.58 2003/10/01 21:30:52 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/executor/nodeSubplan.c,v 1.70.2.4 2010/07/28 04:51:27 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -27,6 +27,7 @@
 #include "utils/array.h"
 #include "utils/datum.h"
 #include "utils/lsyscache.h"
+#include "utils/memutils.h"
 
 
 static Datum ExecHashSubPlan(SubPlanState *node,
@@ -37,7 +38,8 @@ static Datum ExecScanSubPlan(SubPlanState *node,
 				bool *isNull);
 static void buildSubPlanHash(SubPlanState *node);
 static bool findPartialMatch(TupleHashTable hashtable, TupleTableSlot *slot);
-static bool tupleAllNulls(HeapTuple tuple);
+static bool slotAllNulls(TupleTableSlot *slot);
+static bool slotNoNulls(TupleTableSlot *slot);
 
 
 /* ----------------------------------------------------------------
@@ -47,10 +49,16 @@ static bool tupleAllNulls(HeapTuple tuple);
 Datum
 ExecSubPlan(SubPlanState *node,
 			ExprContext *econtext,
-			bool *isNull)
+			bool *isNull,
+			ExprDoneCond *isDone)
 {
 	SubPlan    *subplan = (SubPlan *) node->xprstate.expr;
 
+	/* Set default values for result flags: non-null, not a set result */
+	*isNull = false;
+	if (isDone)
+		*isDone = ExprSingleResult;
+
 	if (subplan->setParam != NIL)
 		elog(ERROR, "cannot set parent params from subquery");
 
@@ -70,24 +78,22 @@ ExecHashSubPlan(SubPlanState *node,
 {
 	SubPlan    *subplan = (SubPlan *) node->xprstate.expr;
 	PlanState  *planstate = node->planstate;
-	ExprContext *innerecontext = node->innerecontext;
 	TupleTableSlot *slot;
-	HeapTuple	tup;
 
 	/* Shouldn't have any direct correlation Vars */
 	if (subplan->parParam != NIL || node->args != NIL)
 		elog(ERROR, "hashed subplan with direct correlation not supported");
 
 	/*
-	 * If first time through or we need to rescan the subplan, build the
-	 * hash table.
+	 * If first time through or we need to rescan the subplan, build the hash
+	 * table.
 	 */
 	if (node->hashtable == NULL || planstate->chgParam != NULL)
 		buildSubPlanHash(node);
 
 	/*
-	 * The result for an empty subplan is always FALSE; no need to
-	 * evaluate lefthand side.
+	 * The result for an empty subplan is always FALSE; no need to evaluate
+	 * lefthand side.
 	 */
 	*isNull = false;
 	if (!node->havehashrows && !node->havenullrows)
@@ -99,39 +105,30 @@ ExecHashSubPlan(SubPlanState *node,
 	 */
 	node->projLeft->pi_exprContext = econtext;
 	slot = ExecProject(node->projLeft, NULL);
-	tup = slot->val;
 
 	/*
-	 * Note: because we are typically called in a per-tuple context, we
-	 * have to explicitly clear the projected tuple before returning.
-	 * Otherwise, we'll have a double-free situation: the per-tuple
-	 * context will probably be reset before we're called again, and then
-	 * the tuple slot will think it still needs to free the tuple.
+	 * Note: because we are typically called in a per-tuple context, we have
+	 * to explicitly clear the projected tuple before returning. Otherwise,
+	 * we'll have a double-free situation: the per-tuple context will probably
+	 * be reset before we're called again, and then the tuple slot will think
+	 * it still needs to free the tuple.
 	 */
 
 	/*
-	 * Since the hashtable routines will use innerecontext's per-tuple
-	 * memory as working memory, be sure to reset it for each tuple.
-	 */
-	ResetExprContext(innerecontext);
-
-	/*
-	 * If the LHS is all non-null, probe for an exact match in the main
-	 * hash table.	If we find one, the result is TRUE. Otherwise, scan
-	 * the partly-null table to see if there are any rows that aren't
-	 * provably unequal to the LHS; if so, the result is UNKNOWN.  (We
-	 * skip that part if we don't care about UNKNOWN.) Otherwise, the
-	 * result is FALSE.
+	 * If the LHS is all non-null, probe for an exact match in the main hash
+	 * table.  If we find one, the result is TRUE. Otherwise, scan the
+	 * partly-null table to see if there are any rows that aren't provably
+	 * unequal to the LHS; if so, the result is UNKNOWN.  (We skip that part
+	 * if we don't care about UNKNOWN.) Otherwise, the result is FALSE.
 	 *
 	 * Note: the reason we can avoid a full scan of the main hash table is
-	 * that the combining operators are assumed never to yield NULL when
-	 * both inputs are non-null.  If they were to do so, we might need to
-	 * produce UNKNOWN instead of FALSE because of an UNKNOWN result in
-	 * comparing the LHS to some main-table entry --- which is a
-	 * comparison we will not even make, unless there's a chance match of
-	 * hash keys.
+	 * that the combining operators are assumed never to yield NULL when both
+	 * inputs are non-null.  If they were to do so, we might need to produce
+	 * UNKNOWN instead of FALSE because of an UNKNOWN result in comparing the
+	 * LHS to some main-table entry --- which is a comparison we will not even
+	 * make, unless there's a chance match of hash keys.
 	 */
-	if (HeapTupleNoNulls(tup))
+	if (slotNoNulls(slot))
 	{
 		if (node->havehashrows &&
 			LookupTupleHashEntry(node->hashtable, slot, NULL) != NULL)
@@ -151,21 +148,21 @@ ExecHashSubPlan(SubPlanState *node,
 	}
 
 	/*
-	 * When the LHS is partly or wholly NULL, we can never return TRUE. If
-	 * we don't care about UNKNOWN, just return FALSE.  Otherwise, if the
-	 * LHS is wholly NULL, immediately return UNKNOWN.	(Since the
-	 * combining operators are strict, the result could only be FALSE if
-	 * the sub-select were empty, but we already handled that case.)
-	 * Otherwise, we must scan both the main and partly-null tables to see
-	 * if there are any rows that aren't provably unequal to the LHS; if
-	 * so, the result is UNKNOWN.  Otherwise, the result is FALSE.
+	 * When the LHS is partly or wholly NULL, we can never return TRUE. If we
+	 * don't care about UNKNOWN, just return FALSE.  Otherwise, if the LHS is
+	 * wholly NULL, immediately return UNKNOWN.  (Since the combining
+	 * operators are strict, the result could only be FALSE if the sub-select
+	 * were empty, but we already handled that case.) Otherwise, we must scan
+	 * both the main and partly-null tables to see if there are any rows that
+	 * aren't provably unequal to the LHS; if so, the result is UNKNOWN.
+	 * Otherwise, the result is FALSE.
 	 */
 	if (node->hashnulls == NULL)
 	{
 		ExecClearTuple(slot);
 		return BoolGetDatum(false);
 	}
-	if (tupleAllNulls(tup))
+	if (slotAllNulls(slot))
 	{
 		ExecClearTuple(slot);
 		*isNull = true;
@@ -206,14 +203,14 @@ ExecScanSubPlan(SubPlanState *node,
 	TupleTableSlot *slot;
 	Datum		result;
 	bool		found = false;	/* TRUE if got at least one subplan tuple */
-	List	   *pvar;
-	List	   *lst;
+	ListCell   *pvar;
+	ListCell   *l;
 	ArrayBuildState *astate = NULL;
 
 	/*
-	 * We are probably in a short-lived expression-evaluation context.
-	 * Switch to the child plan's per-query context for manipulating its
-	 * chgParam, calling ExecProcNode on it, etc.
+	 * We are probably in a short-lived expression-evaluation context. Switch
+	 * to the child plan's per-query context for manipulating its chgParam,
+	 * calling ExecProcNode on it, etc.
 	 */
 	oldcontext = MemoryContextSwitchTo(node->sub_estate->es_query_cxt);
 
@@ -222,43 +219,40 @@ ExecScanSubPlan(SubPlanState *node,
 	 * calculation we have to do is done in the parent econtext, since the
 	 * Param values don't need to have per-query lifetime.)
 	 */
-	pvar = node->args;
-	foreach(lst, subplan->parParam)
+	Assert(list_length(subplan->parParam) == list_length(node->args));
+
+	forboth(l, subplan->parParam, pvar, node->args)
 	{
-		int			paramid = lfirsti(lst);
+		int			paramid = lfirst_int(l);
 		ParamExecData *prm = &(econtext->ecxt_param_exec_vals[paramid]);
 
-		Assert(pvar != NIL);
 		prm->value = ExecEvalExprSwitchContext((ExprState *) lfirst(pvar),
 											   econtext,
 											   &(prm->isnull),
 											   NULL);
-		pvar = lnext(pvar);
 		planstate->chgParam = bms_add_member(planstate->chgParam, paramid);
 	}
-	Assert(pvar == NIL);
 
 	ExecReScan(planstate, NULL);
 
 	/*
-	 * For all sublink types except EXPR_SUBLINK and ARRAY_SUBLINK, the
-	 * result is boolean as are the results of the combining operators. We
-	 * combine results within a tuple (if there are multiple columns)
-	 * using OR semantics if "useOr" is true, AND semantics if not. We
-	 * then combine results across tuples (if the subplan produces more
-	 * than one) using OR semantics for ANY_SUBLINK or AND semantics for
-	 * ALL_SUBLINK. (MULTIEXPR_SUBLINK doesn't allow multiple tuples from
-	 * the subplan.) NULL results from the combining operators are handled
-	 * according to the usual SQL semantics for OR and AND.  The result
-	 * for no input tuples is FALSE for ANY_SUBLINK, TRUE for ALL_SUBLINK,
-	 * NULL for MULTIEXPR_SUBLINK.
+	 * For all sublink types except EXPR_SUBLINK and ARRAY_SUBLINK, the result
+	 * is boolean as are the results of the combining operators. We combine
+	 * results within a tuple (if there are multiple columns) using OR
+	 * semantics if "useOr" is true, AND semantics if not. We then combine
+	 * results across tuples (if the subplan produces more than one) using OR
+	 * semantics for ANY_SUBLINK or AND semantics for ALL_SUBLINK.
+	 * (MULTIEXPR_SUBLINK doesn't allow multiple tuples from the subplan.)
+	 * NULL results from the combining operators are handled according to the
+	 * usual SQL semantics for OR and AND.	The result for no input tuples is
+	 * FALSE for ANY_SUBLINK, TRUE for ALL_SUBLINK, NULL for
+	 * MULTIEXPR_SUBLINK.
 	 *
 	 * For EXPR_SUBLINK we require the subplan to produce no more than one
-	 * tuple, else an error is raised. For ARRAY_SUBLINK we allow the
-	 * subplan to produce more than one tuple. In either case, if zero
-	 * tuples are produced, we return NULL. Assuming we get a tuple, we
-	 * just use its first column (there can be only one non-junk column in
-	 * this case).
+	 * tuple, else an error is raised. For ARRAY_SUBLINK we allow the subplan
+	 * to produce more than one tuple. In either case, if zero tuples are
+	 * produced, we return NULL. Assuming we get a tuple, we just use its
+	 * first column (there can be only one non-junk column in this case).
 	 */
 	result = BoolGetDatum(subLinkType == ALL_SUBLINK);
 	*isNull = false;
@@ -267,12 +261,11 @@ ExecScanSubPlan(SubPlanState *node,
 		 !TupIsNull(slot);
 		 slot = ExecProcNode(planstate))
 	{
-		HeapTuple	tup = slot->val;
-		TupleDesc	tdesc = slot->ttc_tupleDescriptor;
+		TupleDesc	tdesc = slot->tts_tupleDescriptor;
 		Datum		rowresult = BoolGetDatum(!useOr);
 		bool		rownull = false;
 		int			col = 1;
-		List	   *plst;
+		ListCell   *plst;
 
 		if (subLinkType == EXISTS_SUBLINK)
 		{
@@ -291,21 +284,20 @@ ExecScanSubPlan(SubPlanState *node,
 			found = true;
 
 			/*
-			 * We need to copy the subplan's tuple in case the result is
-			 * of pass-by-ref type --- our return value will point into
-			 * this copied tuple!  Can't use the subplan's instance of the
-			 * tuple since it won't still be valid after next
-			 * ExecProcNode() call. node->curTuple keeps track of the
-			 * copied tuple for eventual freeing.
+			 * We need to copy the subplan's tuple in case the result is of
+			 * pass-by-ref type --- our return value will point into this
+			 * copied tuple!  Can't use the subplan's instance of the tuple
+			 * since it won't still be valid after next ExecProcNode() call.
+			 * node->curTuple keeps track of the copied tuple for eventual
+			 * freeing.
 			 */
 			MemoryContextSwitchTo(econtext->ecxt_per_query_memory);
-			tup = heap_copytuple(tup);
 			if (node->curTuple)
 				heap_freetuple(node->curTuple);
-			node->curTuple = tup;
+			node->curTuple = ExecCopySlotTuple(slot);
 			MemoryContextSwitchTo(node->sub_estate->es_query_cxt);
 
-			result = heap_getattr(tup, col, tdesc, isNull);
+			result = heap_getattr(node->curTuple, col, tdesc, isNull);
 			/* keep scanning subplan to make sure there's only one tuple */
 			continue;
 		}
@@ -317,7 +309,7 @@ ExecScanSubPlan(SubPlanState *node,
 
 			found = true;
 			/* stash away current value */
-			dvalue = heap_getattr(tup, 1, tdesc, &disnull);
+			dvalue = slot_getattr(slot, 1, &disnull);
 			astate = accumArrayResult(astate, dvalue, disnull,
 									  tdesc->attrs[0]->atttypid,
 									  oldcontext);
@@ -337,22 +329,22 @@ ExecScanSubPlan(SubPlanState *node,
 		 * For ALL, ANY, and MULTIEXPR sublinks, iterate over combining
 		 * operators for columns of tuple.
 		 */
-		plst = subplan->paramIds;
-		foreach(lst, node->exprs)
+		Assert(list_length(node->exprs) == list_length(subplan->paramIds));
+
+		forboth(l, node->exprs, plst, subplan->paramIds)
 		{
-			ExprState  *exprstate = (ExprState *) lfirst(lst);
-			int			paramid = lfirsti(plst);
+			ExprState  *exprstate = (ExprState *) lfirst(l);
+			int			paramid = lfirst_int(plst);
 			ParamExecData *prmdata;
 			Datum		expresult;
 			bool		expnull;
 
 			/*
-			 * Load up the Param representing this column of the
-			 * sub-select.
+			 * Load up the Param representing this column of the sub-select.
 			 */
 			prmdata = &(econtext->ecxt_param_exec_vals[paramid]);
 			Assert(prmdata->execPlan == NULL);
-			prmdata->value = heap_getattr(tup, col, tdesc,
+			prmdata->value = slot_getattr(slot, col,
 										  &(prmdata->isnull));
 
 			/*
@@ -394,7 +386,6 @@ ExecScanSubPlan(SubPlanState *node,
 				}
 			}
 
-			plst = lnext(plst);
 			col++;
 		}
 
@@ -434,8 +425,8 @@ ExecScanSubPlan(SubPlanState *node,
 	{
 		/*
 		 * deal with empty subplan result.	result/isNull were previously
-		 * initialized correctly for all sublink types except EXPR, ARRAY,
-		 * and MULTIEXPR; for those, return NULL.
+		 * initialized correctly for all sublink types except EXPR, ARRAY, and
+		 * MULTIEXPR; for those, return NULL.
 		 */
 		if (subLinkType == EXPR_SUBLINK ||
 			subLinkType == ARRAY_SUBLINK ||
@@ -465,9 +456,8 @@ buildSubPlanHash(SubPlanState *node)
 {
 	SubPlan    *subplan = (SubPlan *) node->xprstate.expr;
 	PlanState  *planstate = node->planstate;
-	int			ncols = length(node->exprs);
+	int			ncols = list_length(node->exprs);
 	ExprContext *innerecontext = node->innerecontext;
-	MemoryContext tempcxt = innerecontext->ecxt_per_tuple_memory;
 	MemoryContext oldcontext;
 	int			nbuckets;
 	TupleTableSlot *slot;
@@ -476,21 +466,21 @@ buildSubPlanHash(SubPlanState *node)
 	Assert(!subplan->useOr);
 
 	/*
-	 * If we already had any hash tables, destroy 'em; then create empty
-	 * hash table(s).
+	 * If we already had any hash tables, destroy 'em; then create empty hash
+	 * table(s).
 	 *
 	 * If we need to distinguish accurately between FALSE and UNKNOWN (i.e.,
-	 * NULL) results of the IN operation, then we have to store subplan
-	 * output rows that are partly or wholly NULL.	We store such rows in
-	 * a separate hash table that we expect will be much smaller than the
-	 * main table.	(We can use hashing to eliminate partly-null rows that
-	 * are not distinct.  We keep them separate to minimize the cost of
-	 * the inevitable full-table searches; see findPartialMatch.)
+	 * NULL) results of the IN operation, then we have to store subplan output
+	 * rows that are partly or wholly NULL.  We store such rows in a separate
+	 * hash table that we expect will be much smaller than the main table. (We
+	 * can use hashing to eliminate partly-null rows that are not distinct.
+	 * We keep them separate to minimize the cost of the inevitable full-table
+	 * searches; see findPartialMatch.)
 	 *
 	 * If it's not necessary to distinguish FALSE and UNKNOWN, then we don't
 	 * need to store subplan output rows that contain NULL.
 	 */
-	MemoryContextReset(node->tablecxt);
+	MemoryContextReset(node->hashtablecxt);
 	node->hashtable = NULL;
 	node->hashnulls = NULL;
 	node->havehashrows = false;
@@ -506,8 +496,8 @@ buildSubPlanHash(SubPlanState *node)
 										  node->hashfunctions,
 										  nbuckets,
 										  sizeof(TupleHashEntryData),
-										  node->tablecxt,
-										  tempcxt);
+										  node->hashtablecxt,
+										  node->hashtempcxt);
 
 	if (!subplan->unknownEqFalse)
 	{
@@ -525,14 +515,13 @@ buildSubPlanHash(SubPlanState *node)
 											  node->hashfunctions,
 											  nbuckets,
 											  sizeof(TupleHashEntryData),
-											  node->tablecxt,
-											  tempcxt);
+											  node->hashtablecxt,
+											  node->hashtempcxt);
 	}
 
 	/*
-	 * We are probably in a short-lived expression-evaluation context.
-	 * Switch to the child plan's per-query context for calling
-	 * ExecProcNode.
+	 * We are probably in a short-lived expression-evaluation context. Switch
+	 * to the child plan's per-query context for calling ExecProcNode.
 	 */
 	oldcontext = MemoryContextSwitchTo(node->sub_estate->es_query_cxt);
 
@@ -542,45 +531,38 @@ buildSubPlanHash(SubPlanState *node)
 	ExecReScan(planstate, NULL);
 
 	/*
-	 * Scan the subplan and load the hash table(s).  Note that when there
-	 * are duplicate rows coming out of the sub-select, only one copy is
-	 * stored.
+	 * Scan the subplan and load the hash table(s).  Note that when there are
+	 * duplicate rows coming out of the sub-select, only one copy is stored.
 	 */
 	for (slot = ExecProcNode(planstate);
 		 !TupIsNull(slot);
 		 slot = ExecProcNode(planstate))
 	{
-		HeapTuple	tup = slot->val;
-		TupleDesc	tdesc = slot->ttc_tupleDescriptor;
 		int			col = 1;
-		List	   *plst;
+		ListCell   *plst;
 		bool		isnew;
 
 		/*
-		 * Load up the Params representing the raw sub-select outputs,
-		 * then form the projection tuple to store in the hashtable.
+		 * Load up the Params representing the raw sub-select outputs, then
+		 * form the projection tuple to store in the hashtable.
 		 */
 		foreach(plst, subplan->paramIds)
 		{
-			int			paramid = lfirsti(plst);
+			int			paramid = lfirst_int(plst);
 			ParamExecData *prmdata;
 
 			prmdata = &(innerecontext->ecxt_param_exec_vals[paramid]);
 			Assert(prmdata->execPlan == NULL);
-			prmdata->value = heap_getattr(tup, col, tdesc,
+			prmdata->value = slot_getattr(slot, col,
 										  &(prmdata->isnull));
 			col++;
 		}
 		slot = ExecProject(node->projRight, NULL);
-		tup = slot->val;
 
 		/*
 		 * If result contains any nulls, store separately or not at all.
-		 * (Since we know the projection tuple has no junk columns, we can
-		 * just look at the overall hasnull info bit, instead of groveling
-		 * through the columns.)
 		 */
-		if (HeapTupleNoNulls(tup))
+		if (slotNoNulls(slot))
 		{
 			(void) LookupTupleHashEntry(node->hashtable, slot, &isnew);
 			node->havehashrows = true;
@@ -592,17 +574,18 @@ buildSubPlanHash(SubPlanState *node)
 		}
 
 		/*
-		 * Reset innerecontext after each inner tuple to free any memory
-		 * used in hash computation or comparison routines.
+		 * Reset innerecontext after each inner tuple to free any memory used
+		 * during ExecProject.
 		 */
 		ResetExprContext(innerecontext);
 	}
 
 	/*
-	 * Since the projected tuples are in the sub-query's context and not
-	 * the main context, we'd better clear the tuple slot before there's
-	 * any chance of a reset of the sub-query's context.  Else we will
-	 * have the potential for a double free attempt.
+	 * Since the projected tuples are in the sub-query's context and not the
+	 * main context, we'd better clear the tuple slot before there's any
+	 * chance of a reset of the sub-query's context.  Else we will have the
+	 * potential for a double free attempt.  (XXX possibly no longer needed,
+	 * but can't hurt.)
 	 */
 	ExecClearTuple(node->projRight->pi_slot);
 
@@ -622,37 +605,62 @@ findPartialMatch(TupleHashTable hashtable, TupleTableSlot *slot)
 {
 	int			numCols = hashtable->numCols;
 	AttrNumber *keyColIdx = hashtable->keyColIdx;
-	HeapTuple	tuple = slot->val;
-	TupleDesc	tupdesc = slot->ttc_tupleDescriptor;
 	TupleHashIterator hashiter;
 	TupleHashEntry entry;
 
-	ResetTupleHashIterator(hashtable, &hashiter);
+	InitTupleHashIterator(hashtable, &hashiter);
 	while ((entry = ScanTupleHashTable(&hashiter)) != NULL)
 	{
-		if (!execTuplesUnequal(entry->firstTuple,
-							   tuple,
-							   tupdesc,
+		ExecStoreTuple(entry->firstTuple, hashtable->tableslot,
+					   InvalidBuffer, false);
+		if (!execTuplesUnequal(hashtable->tableslot, slot,
 							   numCols, keyColIdx,
 							   hashtable->eqfunctions,
 							   hashtable->tempcxt))
+		{
+			TermTupleHashIterator(&hashiter);
 			return true;
+		}
 	}
+	/* No TermTupleHashIterator call needed here */
 	return false;
 }
 
 /*
- * tupleAllNulls: is the tuple completely NULL?
+ * slotAllNulls: is the slot completely NULL?
+ *
+ * This does not test for dropped columns, which is OK because we only
+ * use it on projected tuples.
  */
 static bool
-tupleAllNulls(HeapTuple tuple)
+slotAllNulls(TupleTableSlot *slot)
 {
-	int			ncols = tuple->t_data->t_natts;
+	int			ncols = slot->tts_tupleDescriptor->natts;
 	int			i;
 
 	for (i = 1; i <= ncols; i++)
 	{
-		if (!heap_attisnull(tuple, i))
+		if (!slot_attisnull(slot, i))
+			return false;
+	}
+	return true;
+}
+
+/*
+ * slotNoNulls: is the slot entirely not NULL?
+ *
+ * This does not test for dropped columns, which is OK because we only
+ * use it on projected tuples.
+ */
+static bool
+slotNoNulls(TupleTableSlot *slot)
+{
+	int			ncols = slot->tts_tupleDescriptor->natts;
+	int			i;
+
+	for (i = 1; i <= ncols; i++)
+	{
+		if (slot_attisnull(slot, i))
 			return false;
 	}
 	return true;
@@ -670,10 +678,9 @@ ExecInitSubPlan(SubPlanState *node, EState *estate)
 	MemoryContext oldcontext;
 
 	/*
-	 * Do access checking on the rangetable entries in the subquery. Here,
-	 * we assume the subquery is a SELECT.
+	 * Do access checking on the rangetable entries in the subquery.
 	 */
-	ExecCheckRTPerms(subplan->rtable, CMD_SELECT);
+	ExecCheckRTPerms(subplan->rtable);
 
 	/*
 	 * initialize my state
@@ -684,7 +691,8 @@ ExecInitSubPlan(SubPlanState *node, EState *estate)
 	node->projRight = NULL;
 	node->hashtable = NULL;
 	node->hashnulls = NULL;
-	node->tablecxt = NULL;
+	node->hashtablecxt = NULL;
+	node->hashtempcxt = NULL;
 	node->innerecontext = NULL;
 	node->keyColIdx = NULL;
 	node->eqfunctions = NULL;
@@ -693,10 +701,10 @@ ExecInitSubPlan(SubPlanState *node, EState *estate)
 	/*
 	 * create an EState for the subplan
 	 *
-	 * The subquery needs its own EState because it has its own rangetable.
-	 * It shares our Param ID space, however.  XXX if rangetable access
-	 * were done differently, the subquery could share our EState, which
-	 * would eliminate some thrashing about in this module...
+	 * The subquery needs its own EState because it has its own rangetable. It
+	 * shares our Param ID space, however.	XXX if rangetable access were done
+	 * differently, the subquery could share our EState, which would eliminate
+	 * some thrashing about in this module...
 	 */
 	sp_estate = CreateExecutorState();
 	node->sub_estate = sp_estate;
@@ -722,21 +730,20 @@ ExecInitSubPlan(SubPlanState *node, EState *estate)
 	MemoryContextSwitchTo(oldcontext);
 
 	/*
-	 * If this plan is un-correlated or undirect correlated one and want
-	 * to set params for parent plan then mark parameters as needing
-	 * evaluation.
+	 * If this plan is un-correlated or undirect correlated one and want to
+	 * set params for parent plan then mark parameters as needing evaluation.
 	 *
 	 * Note that in the case of un-correlated subqueries we don't care about
-	 * setting parent->chgParam here: indices take care about it, for
-	 * others - it doesn't matter...
+	 * setting parent->chgParam here: indices take care about it, for others -
+	 * it doesn't matter...
 	 */
 	if (subplan->setParam != NIL)
 	{
-		List	   *lst;
+		ListCell   *lst;
 
 		foreach(lst, subplan->setParam)
 		{
-			int			paramid = lfirsti(lst);
+			int			paramid = lfirst_int(lst);
 			ParamExecData *prm = &(estate->es_param_exec_vals[paramid]);
 
 			prm->execPlan = node;
@@ -744,8 +751,8 @@ ExecInitSubPlan(SubPlanState *node, EState *estate)
 	}
 
 	/*
-	 * If we are going to hash the subquery output, initialize relevant
-	 * stuff.  (We don't create the hashtable until needed, though.)
+	 * If we are going to hash the subquery output, initialize relevant stuff.
+	 * (We don't create the hashtable until needed, though.)
 	 */
 	if (subplan->useHashTable)
 	{
@@ -757,35 +764,41 @@ ExecInitSubPlan(SubPlanState *node, EState *estate)
 		List	   *lefttlist,
 				   *righttlist,
 				   *leftptlist,
-				   *rightptlist,
-				   *lexpr;
+				   *rightptlist;
+		ListCell   *lexpr;
 
 		/* We need a memory context to hold the hash table(s) */
-		node->tablecxt =
+		node->hashtablecxt =
 			AllocSetContextCreate(CurrentMemoryContext,
 								  "Subplan HashTable Context",
 								  ALLOCSET_DEFAULT_MINSIZE,
 								  ALLOCSET_DEFAULT_INITSIZE,
 								  ALLOCSET_DEFAULT_MAXSIZE);
+		/* and a small one for the hash tables to use as temp storage */
+		node->hashtempcxt =
+			AllocSetContextCreate(CurrentMemoryContext,
+								  "Subplan HashTable Temp Context",
+								  ALLOCSET_SMALL_MINSIZE,
+								  ALLOCSET_SMALL_INITSIZE,
+								  ALLOCSET_SMALL_MAXSIZE);
 		/* and a short-lived exprcontext for function evaluation */
 		node->innerecontext = CreateExprContext(estate);
 		/* Silly little array of column numbers 1..n */
-		ncols = length(node->exprs);
+		ncols = list_length(node->exprs);
 		node->keyColIdx = (AttrNumber *) palloc(ncols * sizeof(AttrNumber));
 		for (i = 0; i < ncols; i++)
 			node->keyColIdx[i] = i + 1;
 
 		/*
 		 * We use ExecProject to evaluate the lefthand and righthand
-		 * expression lists and form tuples.  (You might think that we
-		 * could use the sub-select's output tuples directly, but that is
-		 * not the case if we had to insert any run-time coercions of the
-		 * sub-select's output datatypes; anyway this avoids storing any
-		 * resjunk columns that might be in the sub-select's output.) Run
-		 * through the combining expressions to build tlists for the
-		 * lefthand and righthand sides.  We need both the ExprState list
-		 * (for ExecProject) and the underlying parse Exprs (for
-		 * ExecTypeFromTL).
+		 * expression lists and form tuples.  (You might think that we could
+		 * use the sub-select's output tuples directly, but that is not the
+		 * case if we had to insert any run-time coercions of the sub-select's
+		 * output datatypes; anyway this avoids storing any resjunk columns
+		 * that might be in the sub-select's output.) Run through the
+		 * combining expressions to build tlists for the lefthand and
+		 * righthand sides.  We need both the ExprState list (for ExecProject)
+		 * and the underlying parse Exprs (for ExecTypeFromTL).
 		 *
 		 * We also extract the combining operators themselves to initialize
 		 * the equality and hashing functions for the hash tables.
@@ -807,19 +820,18 @@ ExecInitSubPlan(SubPlanState *node, EState *estate)
 
 			Assert(IsA(fstate, FuncExprState));
 			Assert(IsA(opexpr, OpExpr));
-			Assert(length(fstate->args) == 2);
+			Assert(list_length(fstate->args) == 2);
 
 			/* Process lefthand argument */
-			exstate = (ExprState *) lfirst(fstate->args);
+			exstate = (ExprState *) linitial(fstate->args);
 			expr = exstate->expr;
-			tle = makeTargetEntry(makeResdom(i,
-											 exprType((Node *) expr),
-											 exprTypmod((Node *) expr),
-											 NULL,
-											 false),
-								  expr);
+			tle = makeTargetEntry(expr,
+								  i,
+								  NULL,
+								  false);
 			tlestate = makeNode(GenericExprState);
 			tlestate->xprstate.expr = (Expr *) tle;
+			tlestate->xprstate.evalfunc = NULL;
 			tlestate->arg = exstate;
 			lefttlist = lappend(lefttlist, tlestate);
 			leftptlist = lappend(leftptlist, tle);
@@ -827,14 +839,13 @@ ExecInitSubPlan(SubPlanState *node, EState *estate)
 			/* Process righthand argument */
 			exstate = (ExprState *) lsecond(fstate->args);
 			expr = exstate->expr;
-			tle = makeTargetEntry(makeResdom(i,
-											 exprType((Node *) expr),
-											 exprTypmod((Node *) expr),
-											 NULL,
-											 false),
-								  expr);
+			tle = makeTargetEntry(expr,
+								  i,
+								  NULL,
+								  false);
 			tlestate = makeNode(GenericExprState);
 			tlestate->xprstate.expr = (Expr *) tle;
+			tlestate->xprstate.evalfunc = NULL;
 			tlestate->arg = exstate;
 			righttlist = lappend(righttlist, tlestate);
 			rightptlist = lappend(rightptlist, tle);
@@ -854,35 +865,37 @@ ExecInitSubPlan(SubPlanState *node, EState *estate)
 		}
 
 		/*
-		 * Create a tupletable to hold these tuples.  (Note: we never
-		 * bother to free the tupletable explicitly; that's okay because
-		 * it will never store raw disk tuples that might have associated
-		 * buffer pins.  The only resource involved is memory, which will
-		 * be cleaned up by freeing the query context.)
+		 * Create a tupletable to hold these tuples.  (Note: we never bother
+		 * to free the tupletable explicitly; that's okay because it will
+		 * never store raw disk tuples that might have associated buffer pins.
+		 * The only resource involved is memory, which will be cleaned up by
+		 * freeing the query context.)
 		 */
 		tupTable = ExecCreateTupleTable(2);
 
 		/*
-		 * Construct tupdescs, slots and projection nodes for left and
-		 * right sides.  The lefthand expressions will be evaluated in the
-		 * parent plan node's exprcontext, which we don't have access to
-		 * here.  Fortunately we can just pass NULL for now and fill it in
-		 * later (hack alert!).  The righthand expressions will be
-		 * evaluated in our own innerecontext.
+		 * Construct tupdescs, slots and projection nodes for left and right
+		 * sides.  The lefthand expressions will be evaluated in the parent
+		 * plan node's exprcontext, which we don't have access to here.
+		 * Fortunately we can just pass NULL for now and fill it in later
+		 * (hack alert!).  The righthand expressions will be evaluated in our
+		 * own innerecontext.
 		 */
 		tupDesc = ExecTypeFromTL(leftptlist, false);
 		slot = ExecAllocTableSlot(tupTable);
 		ExecSetSlotDescriptor(slot, tupDesc, true);
 		node->projLeft = ExecBuildProjectionInfo(lefttlist,
 												 NULL,
-												 slot);
+												 slot,
+												 NULL);
 
 		tupDesc = ExecTypeFromTL(rightptlist, false);
 		slot = ExecAllocTableSlot(tupTable);
 		ExecSetSlotDescriptor(slot, tupDesc, true);
 		node->projRight = ExecBuildProjectionInfo(righttlist,
 												  node->innerecontext,
-												  slot);
+												  slot,
+												  NULL);
 	}
 }
 
@@ -907,7 +920,7 @@ ExecSetParamPlan(SubPlanState *node, ExprContext *econtext)
 	SubLinkType subLinkType = subplan->subLinkType;
 	MemoryContext oldcontext;
 	TupleTableSlot *slot;
-	List	   *lst;
+	ListCell   *l;
 	bool		found = false;
 	ArrayBuildState *astate = NULL;
 
@@ -927,14 +940,13 @@ ExecSetParamPlan(SubPlanState *node, ExprContext *econtext)
 		 !TupIsNull(slot);
 		 slot = ExecProcNode(planstate))
 	{
-		HeapTuple	tup = slot->val;
-		TupleDesc	tdesc = slot->ttc_tupleDescriptor;
+		TupleDesc	tdesc = slot->tts_tupleDescriptor;
 		int			i = 1;
 
 		if (subLinkType == EXISTS_SUBLINK)
 		{
 			/* There can be only one param... */
-			int			paramid = lfirsti(subplan->setParam);
+			int			paramid = linitial_int(subplan->setParam);
 			ParamExecData *prm = &(econtext->ecxt_param_exec_vals[paramid]);
 
 			prm->execPlan = NULL;
@@ -951,7 +963,7 @@ ExecSetParamPlan(SubPlanState *node, ExprContext *econtext)
 
 			found = true;
 			/* stash away current value */
-			dvalue = heap_getattr(tup, 1, tdesc, &disnull);
+			dvalue = slot_getattr(slot, 1, &disnull);
 			astate = accumArrayResult(astate, dvalue, disnull,
 									  tdesc->attrs[0]->atttypid,
 									  oldcontext);
@@ -969,29 +981,28 @@ ExecSetParamPlan(SubPlanState *node, ExprContext *econtext)
 		found = true;
 
 		/*
-		 * We need to copy the subplan's tuple into our own context, in
-		 * case any of the params are pass-by-ref type --- the pointers
-		 * stored in the param structs will point at this copied tuple!
-		 * node->curTuple keeps track of the copied tuple for eventual
-		 * freeing.
+		 * We need to copy the subplan's tuple into our own context, in case
+		 * any of the params are pass-by-ref type --- the pointers stored in
+		 * the param structs will point at this copied tuple! node->curTuple
+		 * keeps track of the copied tuple for eventual freeing.
 		 */
 		MemoryContextSwitchTo(econtext->ecxt_per_query_memory);
-		tup = heap_copytuple(tup);
 		if (node->curTuple)
 			heap_freetuple(node->curTuple);
-		node->curTuple = tup;
+		node->curTuple = ExecCopySlotTuple(slot);
 		MemoryContextSwitchTo(node->sub_estate->es_query_cxt);
 
 		/*
 		 * Now set all the setParam params from the columns of the tuple
 		 */
-		foreach(lst, subplan->setParam)
+		foreach(l, subplan->setParam)
 		{
-			int			paramid = lfirsti(lst);
+			int			paramid = lfirst_int(l);
 			ParamExecData *prm = &(econtext->ecxt_param_exec_vals[paramid]);
 
 			prm->execPlan = NULL;
-			prm->value = heap_getattr(tup, i, tdesc, &(prm->isnull));
+			prm->value = heap_getattr(node->curTuple, i, tdesc,
+									  &(prm->isnull));
 			i++;
 		}
 	}
@@ -1001,7 +1012,7 @@ ExecSetParamPlan(SubPlanState *node, ExprContext *econtext)
 		if (subLinkType == EXISTS_SUBLINK)
 		{
 			/* There can be only one param... */
-			int			paramid = lfirsti(subplan->setParam);
+			int			paramid = linitial_int(subplan->setParam);
 			ParamExecData *prm = &(econtext->ecxt_param_exec_vals[paramid]);
 
 			prm->execPlan = NULL;
@@ -1010,9 +1021,9 @@ ExecSetParamPlan(SubPlanState *node, ExprContext *econtext)
 		}
 		else
 		{
-			foreach(lst, subplan->setParam)
+			foreach(l, subplan->setParam)
 			{
-				int			paramid = lfirsti(lst);
+				int			paramid = lfirst_int(l);
 				ParamExecData *prm = &(econtext->ecxt_param_exec_vals[paramid]);
 
 				prm->execPlan = NULL;
@@ -1024,7 +1035,7 @@ ExecSetParamPlan(SubPlanState *node, ExprContext *econtext)
 	else if (subLinkType == ARRAY_SUBLINK)
 	{
 		/* There can be only one param... */
-		int			paramid = lfirsti(subplan->setParam);
+		int			paramid = linitial_int(subplan->setParam);
 		ParamExecData *prm = &(econtext->ecxt_param_exec_vals[paramid]);
 
 		Assert(astate != NULL);
@@ -1067,7 +1078,7 @@ ExecReScanSetParamPlan(SubPlanState *node, PlanState *parent)
 	PlanState  *planstate = node->planstate;
 	SubPlan    *subplan = (SubPlan *) node->xprstate.expr;
 	EState	   *estate = parent->state;
-	List	   *lst;
+	ListCell   *l;
 
 	/* sanity checks */
 	if (subplan->parParam != NIL)
@@ -1084,9 +1095,9 @@ ExecReScanSetParamPlan(SubPlanState *node, PlanState *parent)
 	/*
 	 * Mark this subplan's output parameters as needing recalculation
 	 */
-	foreach(lst, subplan->setParam)
+	foreach(l, subplan->setParam)
 	{
-		int			paramid = lfirsti(lst);
+		int			paramid = lfirst_int(l);
 		ParamExecData *prm = &(estate->es_param_exec_vals[paramid]);
 
 		prm->execPlan = node;
