@@ -1,9 +1,9 @@
 /*
  * psql - the PostgreSQL interactive terminal
  *
- * Copyright (c) 2000-2003, PostgreSQL Global Development Group
+ * Copyright (c) 2000-2005, PostgreSQL Global Development Group
  *
- * $Header: /cvsroot/pgsql/src/bin/psql/print.c,v 1.43 2003/08/14 18:49:42 tgl Exp $
+ * $PostgreSQL: pgsql/src/bin/psql/print.c,v 1.79.2.1 2010/03/01 20:56:15 heikki Exp $
  */
 #include "postgres_fe.h"
 #include "common.h"
@@ -12,7 +12,7 @@
 #include <math.h>
 #include <signal.h>
 
-#if !defined(_MSC_VER) && !defined(__BORLANDC__)
+#ifndef WIN32_CLIENT_ONLY
 #include <unistd.h>
 #endif
 
@@ -24,11 +24,124 @@
 #include <termios.h>
 #endif
 
+#include <locale.h>
+
 #include "pqsignal.h"
 #include "libpq-fe.h"
 
 #include "mbprint.h"
 
+static char *decimal_point;
+static char *grouping;
+static char *thousands_sep;
+
+static void *
+pg_local_malloc(size_t size)
+{
+	void	   *tmp;
+
+	tmp = malloc(size);
+	if (!tmp)
+	{
+		fprintf(stderr, _("out of memory\n"));
+		exit(EXIT_FAILURE);
+	}
+	return tmp;
+}
+
+static int
+integer_digits(const char *my_str)
+{
+	int			frac_len;
+
+	if (my_str[0] == '-')
+		my_str++;
+
+	frac_len = strchr(my_str, '.') ? strlen(strchr(my_str, '.')) : 0;
+
+	return strlen(my_str) - frac_len;
+}
+
+/* Return additional length required for locale-aware numeric output */
+static int
+additional_numeric_locale_len(const char *my_str)
+{
+	int			int_len = integer_digits(my_str),
+				len = 0;
+	int			groupdigits = atoi(grouping);
+
+	if (int_len > 0)
+		/* Don't count a leading separator */
+		len = (int_len / groupdigits - (int_len % groupdigits == 0)) *
+			strlen(thousands_sep);
+
+	if (strchr(my_str, '.') != NULL)
+		len += strlen(decimal_point) - strlen(".");
+
+	return len;
+}
+
+static int
+strlen_with_numeric_locale(const char *my_str)
+{
+	return strlen(my_str) + additional_numeric_locale_len(my_str);
+}
+
+static char *
+format_numeric_locale(const char *my_str)
+{
+	int			i,
+				j,
+				int_len = integer_digits(my_str),
+				leading_digits;
+	int			groupdigits = atoi(grouping);
+	int			new_str_start = 0;
+	char	   *new_str = new_str = pg_local_malloc(
+									 strlen_with_numeric_locale(my_str) + 1);
+
+	leading_digits = (int_len % groupdigits != 0) ?
+		int_len % groupdigits : groupdigits;
+
+	if (my_str[0] == '-')		/* skip over sign, affects grouping
+								 * calculations */
+	{
+		new_str[0] = my_str[0];
+		my_str++;
+		new_str_start = 1;
+	}
+
+	for (i = 0, j = new_str_start;; i++, j++)
+	{
+		/* Hit decimal point? */
+		if (my_str[i] == '.')
+		{
+			strcpy(&new_str[j], decimal_point);
+			j += strlen(decimal_point);
+			/* add fractional part */
+			strcpy(&new_str[j], &my_str[i] + 1);
+			break;
+		}
+
+		/* End of string? */
+		if (my_str[i] == '\0')
+		{
+			new_str[j] = '\0';
+			break;
+		}
+
+		/* Add separator? */
+		if (i != 0 && (i - leading_digits) % groupdigits == 0)
+		{
+			strcpy(&new_str[j], thousands_sep);
+			j += strlen(thousands_sep);
+		}
+
+		new_str[j] = my_str[i];
+	}
+
+	return new_str;
+}
+
 /*************************/
 /* Unaligned text		 */
 /*************************/
@@ -36,8 +149,9 @@
 
 static void
 print_unaligned_text(const char *title, const char *const * headers,
-				  const char *const * cells, const char *const * footers,
- const char *opt_fieldsep, const char *opt_recordsep, bool opt_barebones,
+					 const char *const * cells, const char *const * footers,
+					 const char *opt_align, const char *opt_fieldsep,
+					 const char *opt_recordsep, bool opt_tuples_only,
 					 FILE *fout)
 {
 	unsigned int col_count = 0;
@@ -51,21 +165,21 @@ print_unaligned_text(const char *title, const char *const * headers,
 		opt_recordsep = "";
 
 	/* print title */
-	if (!opt_barebones && title)
+	if (!opt_tuples_only && title)
 		fprintf(fout, "%s%s", title, opt_recordsep);
 
 	/* print headers and count columns */
 	for (ptr = headers; *ptr; ptr++)
 	{
 		col_count++;
-		if (!opt_barebones)
+		if (!opt_tuples_only)
 		{
 			if (col_count > 1)
 				fputs(opt_fieldsep, fout);
 			fputs(*ptr, fout);
 		}
 	}
-	if (!opt_barebones)
+	if (!opt_tuples_only)
 		need_recordsep = true;
 
 	/* print cells */
@@ -78,6 +192,7 @@ print_unaligned_text(const char *title, const char *const * headers,
 			need_recordsep = false;
 		}
 		fputs(*ptr, fout);
+
 		if ((i + 1) % col_count)
 			fputs(opt_fieldsep, fout);
 		else
@@ -87,7 +202,7 @@ print_unaligned_text(const char *title, const char *const * headers,
 
 	/* print footers */
 
-	if (!opt_barebones && footers)
+	if (!opt_tuples_only && footers)
 		for (ptr = footers; *ptr; ptr++)
 		{
 			if (need_recordsep)
@@ -108,9 +223,10 @@ print_unaligned_text(const char *title, const char *const * headers,
 
 static void
 print_unaligned_vertical(const char *title, const char *const * headers,
-				  const char *const * cells, const char *const * footers,
- const char *opt_fieldsep, const char *opt_recordsep, bool opt_barebones,
-						 FILE *fout)
+						 const char *const * cells,
+						 const char *const * footers, const char *opt_align,
+						 const char *opt_fieldsep, const char *opt_recordsep,
+						 bool opt_tuples_only, FILE *fout)
 {
 	unsigned int col_count = 0;
 	unsigned int i;
@@ -122,7 +238,7 @@ print_unaligned_vertical(const char *title, const char *const * headers,
 		opt_recordsep = "";
 
 	/* print title */
-	if (!opt_barebones && title)
+	if (!opt_tuples_only && title)
 		fputs(title, fout);
 
 	/* count columns */
@@ -132,7 +248,7 @@ print_unaligned_vertical(const char *title, const char *const * headers,
 	/* print records */
 	for (i = 0, ptr = cells; *ptr; i++, ptr++)
 	{
-		if (i != 0 || (!opt_barebones && title))
+		if (i != 0 || (!opt_tuples_only && title))
 		{
 			fputs(opt_recordsep, fout);
 			if (i % col_count == 0)
@@ -145,7 +261,7 @@ print_unaligned_vertical(const char *title, const char *const * headers,
 	}
 
 	/* print footers */
-	if (!opt_barebones && footers && *footers)
+	if (!opt_tuples_only && footers && *footers)
 	{
 		fputs(opt_recordsep, fout);
 		for (ptr = footers; *ptr; ptr++)
@@ -204,7 +320,7 @@ _print_horizontal_line(const unsigned int col_count, const unsigned int *widths,
 static void
 print_aligned_text(const char *title, const char *const * headers,
 				   const char *const * cells, const char *const * footers,
-				   const char *opt_align, bool opt_barebones,
+				   const char *opt_align, bool opt_tuples_only,
 				   unsigned short int opt_border, int encoding,
 				   FILE *fout)
 {
@@ -227,14 +343,14 @@ print_aligned_text(const char *title, const char *const * headers,
 		widths = calloc(col_count, sizeof(*widths));
 		if (!widths)
 		{
-			perror("calloc");
+			fprintf(stderr, _("out of memory\n"));
 			exit(EXIT_FAILURE);
 		}
 
 		head_w = calloc(col_count, sizeof(*head_w));
 		if (!head_w)
 		{
-			perror("calloc");
+			fprintf(stderr, _("out of memory\n"));
 			exit(EXIT_FAILURE);
 		}
 	}
@@ -253,7 +369,7 @@ print_aligned_text(const char *title, const char *const * headers,
 		cell_w = calloc(cell_count, sizeof(*cell_w));
 		if (!cell_w)
 		{
-			perror("calloc");
+			fprintf(stderr, _("out of memory\n"));
 			exit(EXIT_FAILURE);
 		}
 	}
@@ -263,7 +379,7 @@ print_aligned_text(const char *title, const char *const * headers,
 	/* calc column widths */
 	for (i = 0; i < col_count; i++)
 	{
-		tmp = pg_wcswidth((unsigned char *) headers[i], strlen(headers[i]), encoding);
+		tmp = pg_wcswidth(headers[i], strlen(headers[i]), encoding);
 		if (tmp > widths[i])
 			widths[i] = tmp;
 		head_w[i] = tmp;
@@ -271,7 +387,7 @@ print_aligned_text(const char *title, const char *const * headers,
 
 	for (i = 0, ptr = cells; *ptr; ptr++, i++)
 	{
-		tmp = pg_wcswidth((unsigned char *) *ptr, strlen(*ptr), encoding);
+		tmp = pg_wcswidth(*ptr, strlen(*ptr), encoding);
 		if (tmp > widths[i % col_count])
 			widths[i % col_count] = tmp;
 		cell_w[i] = tmp;
@@ -288,9 +404,9 @@ print_aligned_text(const char *title, const char *const * headers,
 		total_w += widths[i];
 
 	/* print title */
-	if (title && !opt_barebones)
+	if (title && !opt_tuples_only)
 	{
-		tmp = pg_wcswidth((unsigned char *) title, strlen(title), encoding);
+		tmp = pg_wcswidth(title, strlen(title), encoding);
 		if (tmp >= total_w)
 			fprintf(fout, "%s\n", title);
 		else
@@ -298,7 +414,7 @@ print_aligned_text(const char *title, const char *const * headers,
 	}
 
 	/* print headers */
-	if (!opt_barebones)
+	if (!opt_tuples_only)
 	{
 		if (opt_border == 2)
 			_print_horizontal_line(col_count, widths, opt_border, fout);
@@ -350,10 +466,7 @@ print_aligned_text(const char *title, const char *const * headers,
 
 		/* content */
 		if (opt_align[i % col_count] == 'r')
-		{
-			fprintf(fout, "%*s%s",
-					widths[i % col_count] - cell_w[i], "", cells[i]);
-		}
+			fprintf(fout, "%*s%s", widths[i % col_count] - cell_w[i], "", *ptr);
 		else
 		{
 			if ((i + 1) % col_count == 0 && opt_border != 2)
@@ -384,11 +497,17 @@ print_aligned_text(const char *title, const char *const * headers,
 		_print_horizontal_line(col_count, widths, opt_border, fout);
 
 	/* print footers */
-	if (footers && !opt_barebones)
+	if (footers && !opt_tuples_only)
 		for (ptr = footers; *ptr; ptr++)
 			fprintf(fout, "%s\n", *ptr);
 
+#ifndef __MINGW32__
+
+	/*
+	 * for some reason MinGW outputs an extra newline, so this supresses it
+	 */
 	fputc('\n', fout);
+#endif
 
 	/* clean up */
 	free(cell_w);
@@ -400,8 +519,9 @@ print_aligned_text(const char *title, const char *const * headers,
 
 static void
 print_aligned_vertical(const char *title, const char *const * headers,
-				  const char *const * cells, const char *const * footers,
-					   bool opt_barebones, unsigned short int opt_border,
+					   const char *const * cells, const char *const * footers,
+					   const char *opt_align, bool opt_tuples_only,
+					   unsigned short int opt_border,
 					   int encoding, FILE *fout)
 {
 	unsigned int col_count = 0;
@@ -418,7 +538,7 @@ print_aligned_vertical(const char *title, const char *const * headers,
 
 	if (cells[0] == NULL)
 	{
-		puts(gettext("(No rows)\n"));
+		fprintf(fout, _("(No rows)\n"));
 		return;
 	}
 
@@ -430,7 +550,7 @@ print_aligned_vertical(const char *title, const char *const * headers,
 		head_w = calloc(col_count, sizeof(*head_w));
 		if (!head_w)
 		{
-			perror("calloc");
+			fprintf(stderr, _("out of memory\n"));
 			exit(EXIT_FAILURE);
 		}
 	}
@@ -439,7 +559,7 @@ print_aligned_vertical(const char *title, const char *const * headers,
 
 	for (i = 0; i < col_count; i++)
 	{
-		tmp = pg_wcswidth((unsigned char *) headers[i], strlen(headers[i]), encoding);
+		tmp = pg_wcswidth(headers[i], strlen(headers[i]), encoding);
 		if (tmp > hwidth)
 			hwidth = tmp;
 		head_w[i] = tmp;
@@ -454,7 +574,7 @@ print_aligned_vertical(const char *title, const char *const * headers,
 		cell_w = calloc(cell_count, sizeof(*cell_w));
 		if (!cell_w)
 		{
-			perror("calloc");
+			fprintf(stderr, _("out of memory\n"));
 			exit(EXIT_FAILURE);
 		}
 	}
@@ -464,23 +584,18 @@ print_aligned_vertical(const char *title, const char *const * headers,
 	/* find longest data cell */
 	for (i = 0, ptr = cells; *ptr; ptr++, i++)
 	{
-		tmp = pg_wcswidth((unsigned char *) *ptr, strlen(*ptr), encoding);
+		tmp = pg_wcswidth(*ptr, strlen(*ptr), encoding);
 		if (tmp > dwidth)
 			dwidth = tmp;
 		cell_w[i] = tmp;
 	}
 
 	/* print title */
-	if (!opt_barebones && title)
+	if (!opt_tuples_only && title)
 		fprintf(fout, "%s\n", title);
 
 	/* make horizontal border */
-	divider = malloc(hwidth + dwidth + 10);
-	if (!divider)
-	{
-		perror("malloc");
-		exit(EXIT_FAILURE);
-	}
+	divider = pg_local_malloc(hwidth + dwidth + 10);
 	divider[0] = '\0';
 	if (opt_border == 2)
 		strcat(divider, "+-");
@@ -500,17 +615,11 @@ print_aligned_vertical(const char *title, const char *const * headers,
 	{
 		if (i % col_count == 0)
 		{
-			if (!opt_barebones)
+			if (!opt_tuples_only)
 			{
-				char	   *record_str = malloc(32);
+				char	   *record_str = pg_local_malloc(32);
 				size_t		record_str_len;
 
-				if (!record_str)
-				{
-					perror("malloc");
-					exit(EXIT_FAILURE);
-				}
-
 				if (opt_border == 0)
 					snprintf(record_str, 32, "* Record %d", record++);
 				else
@@ -525,7 +634,7 @@ print_aligned_vertical(const char *title, const char *const * headers,
 
 					if (!div_copy)
 					{
-						perror("malloc");
+						fprintf(stderr, _("out of memory\n"));
 						exit(EXIT_FAILURE);
 					}
 
@@ -560,7 +669,7 @@ print_aligned_vertical(const char *title, const char *const * headers,
 
 	/* print footers */
 
-	if (!opt_barebones && footers && *footers)
+	if (!opt_tuples_only && footers && *footers)
 	{
 		if (opt_border < 2)
 			fputc('\n', fout);
@@ -588,8 +697,10 @@ void
 html_escaped_print(const char *in, FILE *fout)
 {
 	const char *p;
+	bool		leading_space = true;
 
 	for (p = in; *p; p++)
+	{
 		switch (*p)
 		{
 			case '&':
@@ -610,9 +721,19 @@ html_escaped_print(const char *in, FILE *fout)
 			case '\'':
 				fputs("&apos;", fout);
 				break;
+			case ' ':
+				/* protect leading space, for EXPLAIN output */
+				if (leading_space)
+					fputs("&nbsp;", fout);
+				else
+					fputs(" ", fout);
+				break;
 			default:
 				fputc(*p, fout);
 		}
+		if (*p != ' ')
+			leading_space = false;
+	}
 }
 
 
@@ -620,9 +741,9 @@ html_escaped_print(const char *in, FILE *fout)
 static void
 print_html_text(const char *title, const char *const * headers,
 				const char *const * cells, const char *const * footers,
-const char *opt_align, bool opt_barebones, unsigned short int opt_border,
-				const char *opt_table_attr,
-				FILE *fout)
+				const char *opt_align, bool opt_tuples_only,
+				unsigned short int opt_border,
+				const char *opt_table_attr, FILE *fout)
 {
 	unsigned int col_count = 0;
 	unsigned int i;
@@ -634,7 +755,7 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 	fputs(">\n", fout);
 
 	/* print title */
-	if (!opt_barebones && title)
+	if (!opt_tuples_only && title)
 	{
 		fputs("  <caption>", fout);
 		html_escaped_print(title, fout);
@@ -642,19 +763,19 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 	}
 
 	/* print headers and count columns */
-	if (!opt_barebones)
+	if (!opt_tuples_only)
 		fputs("  <tr>\n", fout);
 	for (i = 0, ptr = headers; *ptr; i++, ptr++)
 	{
 		col_count++;
-		if (!opt_barebones)
+		if (!opt_tuples_only)
 		{
 			fputs("    <th align=\"center\">", fout);
 			html_escaped_print(*ptr, fout);
 			fputs("</th>\n", fout);
 		}
 	}
-	if (!opt_barebones)
+	if (!opt_tuples_only)
 		fputs("  </tr>\n", fout);
 
 	/* print cells */
@@ -664,11 +785,12 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 			fputs("  <tr valign=\"top\">\n", fout);
 
 		fprintf(fout, "    <td align=\"%s\">", opt_align[(i) % col_count] == 'r' ? "right" : "left");
-		if ((*ptr)[strspn(*ptr, " \t")] == '\0')		/* is string only
-														 * whitespace? */
+		/* is string only whitespace? */
+		if ((*ptr)[strspn(*ptr, " \t")] == '\0')
 			fputs("&nbsp; ", fout);
 		else
 			html_escaped_print(*ptr, fout);
+
 		fputs("</td>\n", fout);
 
 		if ((i + 1) % col_count == 0)
@@ -679,7 +801,7 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 
 	/* print footers */
 
-	if (!opt_barebones && footers && *footers)
+	if (!opt_tuples_only && footers && *footers)
 	{
 		fputs("<p>", fout);
 		for (ptr = footers; *ptr; ptr++)
@@ -696,10 +818,10 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 
 static void
 print_html_vertical(const char *title, const char *const * headers,
-				  const char *const * cells, const char *const * footers,
-const char *opt_align, bool opt_barebones, unsigned short int opt_border,
-					const char *opt_table_attr,
-					FILE *fout)
+					const char *const * cells, const char *const * footers,
+					const char *opt_align, bool opt_tuples_only,
+					unsigned short int opt_border,
+					const char *opt_table_attr, FILE *fout)
 {
 	unsigned int col_count = 0;
 	unsigned int i;
@@ -712,7 +834,7 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 	fputs(">\n", fout);
 
 	/* print title */
-	if (!opt_barebones && title)
+	if (!opt_tuples_only && title)
 	{
 		fputs("  <caption>", fout);
 		html_escaped_print(title, fout);
@@ -728,7 +850,7 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 	{
 		if (i % col_count == 0)
 		{
-			if (!opt_barebones)
+			if (!opt_tuples_only)
 				fprintf(fout, "\n  <tr><td colspan=\"2\" align=\"center\">Record %d</td></tr>\n", record++);
 			else
 				fputs("\n  <tr><td colspan=\"2\">&nbsp;</td></tr>\n", fout);
@@ -739,18 +861,19 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 		fputs("</th>\n", fout);
 
 		fprintf(fout, "    <td align=\"%s\">", opt_align[i % col_count] == 'r' ? "right" : "left");
-		if ((*ptr)[strspn(*ptr, " \t")] == '\0')		/* is string only
-														 * whitespace? */
+		/* is string only whitespace? */
+		if ((*ptr)[strspn(*ptr, " \t")] == '\0')
 			fputs("&nbsp; ", fout);
 		else
 			html_escaped_print(*ptr, fout);
+
 		fputs("</td>\n  </tr>\n", fout);
 	}
 
 	fputs("</table>\n", fout);
 
 	/* print footers */
-	if (!opt_barebones && footers && *footers)
+	if (!opt_tuples_only && footers && *footers)
 	{
 		fputs("<p>", fout);
 		for (ptr = footers; *ptr; ptr++)
@@ -766,7 +889,7 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 
 
 /*************************/
-/* LaTeX				 */
+/* LaTeX		 */
 /*************************/
 
 
@@ -787,6 +910,9 @@ latex_escaped_print(const char *in, FILE *fout)
 			case '$':
 				fputs("\\$", fout);
 				break;
+			case '_':
+				fputs("\\_", fout);
+				break;
 			case '{':
 				fputs("\\{", fout);
 				break;
@@ -809,63 +935,60 @@ latex_escaped_print(const char *in, FILE *fout)
 static void
 print_latex_text(const char *title, const char *const * headers,
 				 const char *const * cells, const char *const * footers,
-const char *opt_align, bool opt_barebones, unsigned short int opt_border,
+				 const char *opt_align, bool opt_tuples_only,
+				 unsigned short int opt_border,
 				 FILE *fout)
 {
 	unsigned int col_count = 0;
 	unsigned int i;
-	const char *cp;
 	const char *const * ptr;
 
 
 	/* print title */
-	if (!opt_barebones && title)
+	if (!opt_tuples_only && title)
 	{
 		fputs("\\begin{center}\n", fout);
 		latex_escaped_print(title, fout);
 		fputs("\n\\end{center}\n\n", fout);
 	}
 
+	/* count columns */
+	for (ptr = headers; *ptr; ptr++)
+		col_count++;
+
 	/* begin environment and set alignments and borders */
 	fputs("\\begin{tabular}{", fout);
-	if (opt_border == 0)
-		fputs(opt_align, fout);
-	else if (opt_border == 1)
+
+	if (opt_border == 2)
+		fputs("| ", fout);
+	for (i = 0; i < col_count; i++)
 	{
-		for (cp = opt_align; *cp; cp++)
-		{
-			if (cp != opt_align)
-				fputc('|', fout);
-			fputc(*cp, fout);
-		}
-	}
-	else if (opt_border == 2)
-	{
-		for (cp = opt_align; *cp; cp++)
-		{
-			fputc('|', fout);
-			fputc(*cp, fout);
-		}
-		fputc('|', fout);
+		fputc(*(opt_align + i), fout);
+		if (opt_border != 0 && i < col_count - 1)
+			fputs(" | ", fout);
 	}
+	if (opt_border == 2)
+		fputs(" |", fout);
+
 	fputs("}\n", fout);
 
-	if (!opt_barebones && opt_border == 2)
+	if (!opt_tuples_only && opt_border == 2)
 		fputs("\\hline\n", fout);
 
 	/* print headers and count columns */
-	for (i = 0, ptr = headers; *ptr; i++, ptr++)
+	for (i = 0, ptr = headers; i < col_count; i++, ptr++)
 	{
-		col_count++;
-		if (!opt_barebones)
+		if (!opt_tuples_only)
 		{
 			if (i != 0)
 				fputs(" & ", fout);
+			fputs("\\textit{", fout);
 			latex_escaped_print(*ptr, fout);
+			fputc('}', fout);
 		}
 	}
 
-	if (!opt_barebones)
+	if (!opt_tuples_only)
 	{
 		fputs(" \\\\\n", fout);
 		fputs("\\hline\n", fout);
@@ -885,12 +1008,12 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 	if (opt_border == 2)
 		fputs("\\hline\n", fout);
 
-	fputs("\\end{tabular}\n\n", fout);
+	fputs("\\end{tabular}\n\n\\noindent ", fout);
 
 
 	/* print footers */
 
-	if (footers && !opt_barebones)
+	if (footers && !opt_tuples_only)
 		for (ptr = footers; *ptr; ptr++)
 		{
 			latex_escaped_print(*ptr, fout);
@@ -904,8 +1027,9 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 
 static void
 print_latex_vertical(const char *title, const char *const * headers,
-				  const char *const * cells, const char *const * footers,
-const char *opt_align, bool opt_barebones, unsigned short int opt_border,
+					 const char *const * cells, const char *const * footers,
+					 const char *opt_align, bool opt_tuples_only,
+					 unsigned short int opt_border,
 					 FILE *fout)
 {
 	unsigned int col_count = 0;
@@ -916,7 +1040,7 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 	(void) opt_align;			/* currently unused parameter */
 
 	/* print title */
-	if (!opt_barebones && title)
+	if (!opt_tuples_only && title)
 	{
 		fputs("\\begin{center}\n", fout);
 		latex_escaped_print(title, fout);
@@ -945,11 +1069,15 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 		/* new record */
 		if (i % col_count == 0)
 		{
-			if (!opt_barebones)
+			if (!opt_tuples_only)
 			{
 				if (opt_border == 2)
+				{
 					fputs("\\hline\n", fout);
-				fprintf(fout, "\\multicolumn{2}{c}{Record %d} \\\\\n", record++);
+					fprintf(fout, "\\multicolumn{2}{|c|}{\\textit{Record %d}} \\\\\n", record++);
+				}
+				else
+					fprintf(fout, "\\multicolumn{2}{c}{\\textit{Record %d}} \\\\\n", record++);
 			}
 			if (opt_border >= 1)
 				fputs("\\hline\n", fout);
@@ -964,12 +1092,12 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 	if (opt_border == 2)
 		fputs("\\hline\n", fout);
 
-	fputs("\\end{tabular}\n\n", fout);
+	fputs("\\end{tabular}\n\n\\noindent ", fout);
 
 
 	/* print footers */
 
-	if (footers && !opt_barebones)
+	if (footers && !opt_tuples_only)
 		for (ptr = footers; *ptr; ptr++)
 		{
 			latex_escaped_print(*ptr, fout);
@@ -981,6 +1109,211 @@ const char *opt_align, bool opt_barebones, unsigned short int opt_border,
 
 
 
+/*************************/
+/* Troff -ms		 */
+/*************************/
+
+
+static void
+troff_ms_escaped_print(const char *in, FILE *fout)
+{
+	const char *p;
+
+	for (p = in; *p; p++)
+		switch (*p)
+		{
+			case '\\':
+				fputs("\\(rs", fout);
+				break;
+			default:
+				fputc(*p, fout);
+		}
+}
+
+
+
+static void
+print_troff_ms_text(const char *title, const char *const * headers,
+					const char *const * cells, const char *const * footers,
+					const char *opt_align, bool opt_tuples_only,
+					unsigned short int opt_border,
+					FILE *fout)
+{
+	unsigned int col_count = 0;
+	unsigned int i;
+	const char *const * ptr;
+
+
+	/* print title */
+	if (!opt_tuples_only && title)
+	{
+		fputs(".LP\n.DS C\n", fout);
+		troff_ms_escaped_print(title, fout);
+		fputs("\n.DE\n", fout);
+	}
+
+	/* count columns */
+	for (ptr = headers; *ptr; ptr++)
+		col_count++;
+
+	/* begin environment and set alignments and borders */
+	fputs(".LP\n.TS\n", fout);
+	if (opt_border == 2)
+		fputs("center box;\n", fout);
+	else
+		fputs("center;\n", fout);
+
+	for (i = 0; i < col_count; i++)
+	{
+		fputc(*(opt_align + i), fout);
+		if (opt_border > 0 && i < col_count - 1)
+			fputs(" | ", fout);
+	}
+	fputs(".\n", fout);
+
+	/* print headers and count columns */
+	for (i = 0, ptr = headers; i < col_count; i++, ptr++)
+	{
+		if (!opt_tuples_only)
+		{
+			if (i != 0)
+				fputc('\t', fout);
+			fputs("\\fI", fout);
+			troff_ms_escaped_print(*ptr, fout);
+			fputs("\\fP", fout);
+		}
+	}
+
+	if (!opt_tuples_only)
+		fputs("\n_\n", fout);
+
+	/* print cells */
+	for (i = 0, ptr = cells; *ptr; i++, ptr++)
+	{
+		troff_ms_escaped_print(*ptr, fout);
+
+		if ((i + 1) % col_count == 0)
+			fputc('\n', fout);
+		else
+			fputc('\t', fout);
+	}
+
+	fputs(".TE\n.DS L\n", fout);
+
+
+	/* print footers */
+
+	if (footers && !opt_tuples_only)
+		for (ptr = footers; *ptr; ptr++)
+		{
+			troff_ms_escaped_print(*ptr, fout);
+			fputc('\n', fout);
+		}
+
+	fputs(".DE\n", fout);
+}
+
+
+
+static void
+print_troff_ms_vertical(const char *title, const char *const * headers,
+					  const char *const * cells, const char *const * footers,
+						const char *opt_align, bool opt_tuples_only,
+						unsigned short int opt_border,
+						FILE *fout)
+{
+	unsigned int col_count = 0;
+	unsigned int i;
+	const char *const * ptr;
+	unsigned int record = 1;
+	unsigned short current_format = 0;	/* 0=none, 1=header, 2=body */
+
+	(void) opt_align;			/* currently unused parameter */
+
+	/* print title */
+	if (!opt_tuples_only && title)
+	{
+		fputs(".LP\n.DS C\n", fout);
+		troff_ms_escaped_print(title, fout);
+		fputs("\n.DE\n", fout);
+	}
+
+	/* begin environment and set alignments and borders */
+	fputs(".LP\n.TS\n", fout);
+	if (opt_border == 2)
+		fputs("center box;\n", fout);
+	else
+		fputs("center;\n", fout);
+
+	/* basic format */
+	if (opt_tuples_only)
+		fputs("c l;\n", fout);
+
+	/* count columns */
+	for (ptr = headers; *ptr; ptr++)
+		col_count++;
+
+	/* print records */
+	for (i = 0, ptr = cells; *ptr; i++, ptr++)
+	{
+		/* new record */
+		if (i % col_count == 0)
+		{
+			if (!opt_tuples_only)
+			{
+				if (current_format != 1)
+				{
+					if (opt_border == 2 && i > 0)
+						fputs("_\n", fout);
+					if (current_format != 0)
+						fputs(".T&\n", fout);
+					fputs("c s.\n", fout);
+					current_format = 1;
+				}
+				fprintf(fout, "\\fIRecord %d\\fP\n", record++);
+			}
+			if (opt_border >= 1)
+				fputs("_\n", fout);
+		}
+
+		if (!opt_tuples_only)
+		{
+			if (current_format != 2)
+			{
+				if (current_format != 0)
+					fputs(".T&\n", fout);
+				if (opt_border != 1)
+					fputs("c l.\n", fout);
+				else
+					fputs("c | l.\n", fout);
+				current_format = 2;
+			}
+		}
+
+		troff_ms_escaped_print(headers[i % col_count], fout);
+		fputc('\t', fout);
+		troff_ms_escaped_print(*ptr, fout);
+
+		fputc('\n', fout);
+	}
+
+	fputs(".TE\n.DS L\n", fout);
+
+
+	/* print footers */
+
+	if (footers && !opt_tuples_only)
+		for (ptr = footers; *ptr; ptr++)
+		{
+			troff_ms_escaped_print(*ptr, fout);
+			fputc('\n', fout);
+		}
+
+	fputs(".DE\n", fout);
+}
+
+
+
 /********************************/
 /* Public functions		*/
 /********************************/
@@ -1038,11 +1371,12 @@ printTable(const char *title,
 		   const char *const * cells,
 		   const char *const * footers,
 		   const char *align,
-		   const printTableOpt *opt, FILE *fout)
+		   const printTableOpt *opt, FILE *fout, FILE *flog)
 {
 	const char *default_footer[] = {NULL};
 	unsigned short int border = opt->border;
 	FILE	   *output;
+	bool		use_expanded;
 
 	if (opt->format == PRINT_NOTHING)
 		return;
@@ -1053,6 +1387,16 @@ printTable(const char *title,
 	if (opt->format != PRINT_HTML && border > 2)
 		border = 2;
 
+	/*
+	 * We only want to display the results in "expanded" format if this is a
+	 * normal (user-submitted) query, not a table we're printing for a slash
+	 * command.
+	 */
+	if (opt->expanded)
+		use_expanded = true;
+	else
+		use_expanded = false;
+
 	if (fout == stdout)
 	{
 		int			col_count = 0,
@@ -1084,34 +1428,64 @@ printTable(const char *title,
 
 	/* print the stuff */
 
+	if (flog)
+		print_aligned_text(title, headers, cells, footers, align, opt->tuples_only, border, opt->encoding, flog);
+
 	switch (opt->format)
 	{
 		case PRINT_UNALIGNED:
-			if (opt->expanded)
-				print_unaligned_vertical(title, headers, cells, footers, opt->fieldSep, opt->recordSep, opt->tuples_only, output);
+			if (use_expanded)
+				print_unaligned_vertical(title, headers, cells, footers, align,
+										 opt->fieldSep, opt->recordSep,
+										 opt->tuples_only, output);
 			else
-				print_unaligned_text(title, headers, cells, footers, opt->fieldSep, opt->recordSep, opt->tuples_only, output);
+				print_unaligned_text(title, headers, cells, footers, align,
+									 opt->fieldSep, opt->recordSep,
+									 opt->tuples_only, output);
 			break;
 		case PRINT_ALIGNED:
-			if (opt->expanded)
-				print_aligned_vertical(title, headers, cells, footers, opt->tuples_only, border, opt->encoding, output);
+			if (use_expanded)
+				print_aligned_vertical(title, headers, cells, footers, align,
+									   opt->tuples_only, border,
+									   opt->encoding, output);
 			else
-				print_aligned_text(title, headers, cells, footers, align, opt->tuples_only, border, opt->encoding, output);
+				print_aligned_text(title, headers, cells, footers, align,
+								   opt->tuples_only,
+								   border, opt->encoding, output);
 			break;
 		case PRINT_HTML:
-			if (opt->expanded)
-				print_html_vertical(title, headers, cells, footers, align, opt->tuples_only, border, opt->tableAttr, output);
+			if (use_expanded)
+				print_html_vertical(title, headers, cells, footers, align,
+									opt->tuples_only,
+									border, opt->tableAttr, output);
 			else
-				print_html_text(title, headers, cells, footers, align, opt->tuples_only, border, opt->tableAttr, output);
+				print_html_text(title, headers, cells, footers,
+						 align, opt->tuples_only, border,
+								opt->tableAttr, output);
 			break;
 		case PRINT_LATEX:
-			if (opt->expanded)
-				print_latex_vertical(title, headers, cells, footers, align, opt->tuples_only, border, output);
+			if (use_expanded)
+				print_latex_vertical(title, headers, cells, footers, align,
+									 opt->tuples_only,
+									 border, output);
 			else
-				print_latex_text(title, headers, cells, footers, align, opt->tuples_only, border, output);
+				print_latex_text(title, headers, cells, footers, align,
+								 opt->tuples_only,
+								 border, output);
+			break;
+		case PRINT_TROFF_MS:
+			if (use_expanded)
+				print_troff_ms_vertical(title, headers, cells, footers, align,
+										opt->tuples_only,
+										border, output);
+			else
+				print_troff_ms_text(title, headers, cells, footers, align,
+									opt->tuples_only,
+									border, output);
 			break;
 		default:
-			fprintf(stderr, "+ Oops, you shouldn't see this!\n");
+			fprintf(stderr, _("invalid output format (internal error): %d"), opt->format);
+			exit(EXIT_FAILURE);
 	}
 
 	/* Only close if we used the pager */
@@ -1127,45 +1501,94 @@ printTable(const char *title,
 
 
 void
-printQuery(const PGresult *result, const printQueryOpt *opt, FILE *fout)
+printQuery(const PGresult *result, const printQueryOpt *opt, FILE *fout, FILE *flog)
 {
 	int			nfields;
+	int			ncells;
 	const char **headers;
 	const char **cells;
+	bool	   *cellmustfree = NULL;
 	char	  **footers;
 	char	   *align;
 	int			i;
 
-
 	/* extract headers */
-
 	nfields = PQnfields(result);
 
 	headers = calloc(nfields + 1, sizeof(*headers));
 	if (!headers)
 	{
-		perror("calloc");
+		fprintf(stderr, _("out of memory\n"));
 		exit(EXIT_FAILURE);
 	}
 
 	for (i = 0; i < nfields; i++)
 		headers[i] = mbvalidate(PQfname(result, i), opt->topt.encoding);
 
-	/* set cells */
-
-	cells = calloc(nfields * PQntuples(result) + 1, sizeof(*cells));
-	if (!cells)
+	/* set alignment */
+	align = calloc(nfields + 1, sizeof(*align));
+	if (!align)
 	{
-		perror("calloc");
+		fprintf(stderr, _("out of memory\n"));
 		exit(EXIT_FAILURE);
 	}
 
-	for (i = 0; i < nfields * PQntuples(result); i++)
+	for (i = 0; i < nfields; i++)
+	{
+		Oid			ftype = PQftype(result, i);
+
+		if (ftype == 20 ||		/* int8 */
+			ftype == 21 ||		/* int2 */
+			ftype == 23 ||		/* int4 */
+			(ftype >= 26 && ftype <= 30) ||		/* ?id */
+			ftype == 700 ||		/* float4 */
+			ftype == 701 ||		/* float8 */
+			ftype == 790 ||		/* money */
+			ftype == 1700		/* numeric */
+			)
+			align[i] = 'r';
+		else
+			align[i] = 'l';
+	}
+
+	/* set cells */
+	ncells = PQntuples(result) * nfields;
+	cells = calloc(ncells + 1, sizeof(*cells));
+	if (!cells)
+	{
+		fprintf(stderr, _("out of memory\n"));
+		exit(EXIT_FAILURE);
+	}
+
+	for (i = 0; i < ncells; i++)
 	{
 		if (PQgetisnull(result, i / nfields, i % nfields))
 			cells[i] = opt->nullPrint ? opt->nullPrint : "";
 		else
+		{
 			cells[i] = mbvalidate(PQgetvalue(result, i / nfields, i % nfields), opt->topt.encoding);
+
+			if (align[i % nfields] == 'r' && opt->topt.numericLocale)
+			{
+				cells[i] = format_numeric_locale(cells[i]);
+
+				/*
+				 * format_numeric_locale() returns a malloc'd string, 
+				 * remember that it needs to be freed.
+				 */
+				if (cellmustfree == NULL)
+				{
+					cellmustfree = calloc(ncells + 1, sizeof(bool));
+					if (!cellmustfree)
+					{
+						fprintf(stderr, _("out of memory\n"));
+						exit(EXIT_FAILURE);
+					}
+				}
+
+				cellmustfree[i] = true;
+			}
+		}
 	}
 
 	/* set footers */
@@ -1177,25 +1600,24 @@ printQuery(const PGresult *result, const printQueryOpt *opt, FILE *fout)
 		footers = calloc(2, sizeof(*footers));
 		if (!footers)
 		{
-			perror("calloc");
+			fprintf(stderr, _("out of memory\n"));
 			exit(EXIT_FAILURE);
 		}
 
-		footers[0] = malloc(100);
+		footers[0] = pg_local_malloc(100);
 		if (PQntuples(result) == 1)
-			snprintf(footers[0], 100, gettext("(1 row)"));
+			snprintf(footers[0], 100, _("(1 row)"));
 		else
-			snprintf(footers[0], 100, gettext("(%d rows)"), PQntuples(result));
+			snprintf(footers[0], 100, _("(%d rows)"), PQntuples(result));
 	}
 	else
 		footers = NULL;
 
 	/* set alignment */
-
 	align = calloc(nfields + 1, sizeof(*align));
 	if (!align)
 	{
-		perror("calloc");
+		fprintf(stderr, _("out of memory\n"));
 		exit(EXIT_FAILURE);
 	}
 
@@ -1218,13 +1640,21 @@ printQuery(const PGresult *result, const printQueryOpt *opt, FILE *fout)
 	}
 
 	/* call table printer */
-
 	printTable(opt->title, headers, cells,
-			   footers ? (const char *const *) footers : (const char *const *) (opt->footers),
-			   align, &opt->topt, fout);
+			   (const char *const *) footers,
+			   align, &opt->topt, fout, flog);
 
-	free((void *) headers);
-	free((void *) cells);
+	free(headers);
+	if (cellmustfree)
+	{
+		for (i = 0; i < ncells; i++)
+		{
+			if (cellmustfree[i])
+				free((char *) cells[i]);
+		}
+		free(cellmustfree);
+	}
+	free(cells);
 	if (footers)
 	{
 		free(footers[0]);
@@ -1234,4 +1664,25 @@ printQuery(const PGresult *result, const printQueryOpt *opt, FILE *fout)
 }
 
 
-/* the end */
+void
+setDecimalLocale(void)
+{
+	struct lconv *extlconv;
+
+	extlconv = localeconv();
+
+	if (*extlconv->decimal_point)
+		decimal_point = strdup(extlconv->decimal_point);
+	else
+		decimal_point = ".";	/* SQL output standard */
+	if (*extlconv->grouping && atoi(extlconv->grouping) > 0)
+		grouping = strdup(extlconv->grouping);
+	else
+		grouping = "3";			/* most common */
+	if (*extlconv->thousands_sep)
+		thousands_sep = strdup(extlconv->thousands_sep);
+	else if (*decimal_point != ',')
+		thousands_sep = ",";
+	else
+		thousands_sep = ".";
+}
