@@ -6,10 +6,10 @@
  *	   including abstime, reltime, date, and time.
  *
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
- * $Id: datetime.h,v 1.45.2.1 2005/05/26 02:14:32 neilc Exp $
+ * $PostgreSQL: pgsql/src/include/utils/datetime.h,v 1.57.2.1 2007/05/29 04:59:44 neilc Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -18,7 +18,6 @@
 
 #include <limits.h>
 #include <math.h>
-#include <time.h>
 
 #include "utils/timestamp.h"
 
@@ -100,17 +99,19 @@
 #define HOUR	10
 #define MINUTE	11
 #define SECOND	12
-#define DOY		13
-#define DOW		14
-#define UNITS	15
-#define ADBC	16
+#define MILLISECOND 13
+#define MICROSECOND 14
+#define DOY		15
+#define DOW		16
+#define UNITS	17
+#define ADBC	18
 /* these are only for relative dates */
-#define AGO		17
-#define ABS_BEFORE		18
-#define ABS_AFTER		19
+#define AGO		19
+#define ABS_BEFORE		20
+#define ABS_AFTER		21
 /* generic fields to help with parsing */
-#define ISODATE 20
-#define ISOTIME 21
+#define ISODATE 22
+#define ISOTIME 23
 /* reserved for unrecognized string values */
 #define UNKNOWN_FIELD	31
 
@@ -173,13 +174,15 @@
 
 #define DTK_M(t)		(0x01 << (t))
 
+/* Convenvience: a second, plus any fractional component */
+#define DTK_ALL_SECS_M	(DTK_M(SECOND) | DTK_M(MILLISECOND) | DTK_M(MICROSECOND))
 #define DTK_DATE_M		(DTK_M(YEAR) | DTK_M(MONTH) | DTK_M(DAY))
-#define DTK_TIME_M		(DTK_M(HOUR) | DTK_M(MINUTE) | DTK_M(SECOND))
+#define DTK_TIME_M		(DTK_M(HOUR) | DTK_M(MINUTE) | DTK_ALL_SECS_M)
 
-#define MAXDATELEN		51		/* maximum possible length of an input
-								 * date string (not counting tr. null) */
-#define MAXDATEFIELDS	25		/* maximum possible number of fields in a
-								 * date string */
+#define MAXDATELEN		51		/* maximum possible length of an input date
+								 * string (not counting tr. null) */
+#define MAXDATEFIELDS	25		/* maximum possible number of fields in a date
+								 * string */
 #define TOKMAXLEN		10		/* only this many chars are stored in
 								 * datetktbl */
 
@@ -204,8 +207,8 @@ typedef struct
  */
 #define FMODULO(t,q,u) \
 do { \
-	q = ((t < 0) ? ceil(t / u) : floor(t / u)); \
-	if (q != 0) t -= rint(q * u); \
+	(q) = (((t) < 0) ? ceil((t) / (u)) : floor((t) / (u))); \
+	if ((q) != 0) (t) -= rint((q) * (u)); \
 } while(0)
 
 /* TMODULO()
@@ -216,14 +219,14 @@ do { \
 #ifdef HAVE_INT64_TIMESTAMP
 #define TMODULO(t,q,u) \
 do { \
-	q = (t / u); \
-	if (q != 0) t -= (q * u); \
+	(q) = ((t) / (u)); \
+	if ((q) != 0) (t) -= ((q) * (u)); \
 } while(0)
 #else
 #define TMODULO(t,q,u) \
 do { \
-	q = ((t < 0) ? ceil(t / u) : floor(t / u)); \
-	if (q != 0) t -= rint(q * u); \
+	(q) = (((t) < 0) ? ceil((t) / (u)) : floor((t) / (u))); \
+	if ((q) != 0) (t) -= rint((q) * (u)); \
 } while(0)
 #endif
 
@@ -232,10 +235,11 @@ do { \
  * Include check for leap year.
  */
 
-extern int	day_tab[2][13];
+extern const int day_tab[2][13];
 
 #define isleap(y) (((y) % 4) == 0 && (((y) % 100) != 0 || ((y) % 400) == 0))
 
+
 /* Julian date support for date2j() and j2date()
  *
  * IS_VALID_JULIAN checks the minimum date exactly, but is a bit sloppy
@@ -257,40 +261,22 @@ extern int	day_tab[2][13];
 #define UNIX_EPOCH_JDATE		2440588 /* == date2j(1970, 1, 1) */
 #define POSTGRES_EPOCH_JDATE	2451545 /* == date2j(2000, 1, 1) */
 
-/*
- * Info about limits of the Unix time_t data type.  We assume that time_t
- * is a signed int32 with origin 1970-01-01.  Note this is only relevant
- * when we use the C library's time routines for timezone processing.
- */
-#define UTIME_MINYEAR (1901)
-#define UTIME_MINMONTH (12)
-#define UTIME_MINDAY (14)
-#define UTIME_MAXYEAR (2038)
-#define UTIME_MAXMONTH (01)
-#define UTIME_MAXDAY (18)
-
-#define IS_VALID_UTIME(y,m,d) ((((y) > UTIME_MINYEAR) \
- || (((y) == UTIME_MINYEAR) && (((m) > UTIME_MINMONTH) \
-  || (((m) == UTIME_MINMONTH) && ((d) >= UTIME_MINDAY))))) \
- && (((y) < UTIME_MAXYEAR) \
- || (((y) == UTIME_MAXYEAR) && (((m) < UTIME_MAXMONTH) \
-  || (((m) == UTIME_MAXMONTH) && ((d) <= UTIME_MAXDAY))))))
 
 /*
  * Datetime input parsing routines (ParseDateTime, DecodeDateTime, etc)
- * return zero or a positive value on success.  On failure, they return
+ * return zero or a positive value on success.	On failure, they return
  * one of these negative code values.  DateTimeParseError may be used to
  * produce a correct ereport.
  */
 #define DTERR_BAD_FORMAT		(-1)
 #define DTERR_FIELD_OVERFLOW	(-2)
-#define DTERR_MD_FIELD_OVERFLOW	(-3)	/* triggers hint about DateStyle */
-#define DTERR_INTERVAL_OVERFLOW	(-4)
+#define DTERR_MD_FIELD_OVERFLOW (-3)	/* triggers hint about DateStyle */
+#define DTERR_INTERVAL_OVERFLOW (-4)
 #define DTERR_TZDISP_OVERFLOW	(-5)
 
 
-extern void GetCurrentDateTime(struct tm * tm);
-extern void GetCurrentTimeUsec(struct tm * tm, fsec_t *fsec, int *tzp);
+extern void GetCurrentDateTime(struct pg_tm * tm);
+extern void GetCurrentTimeUsec(struct pg_tm * tm, fsec_t *fsec, int *tzp);
 extern void j2date(int jd, int *year, int *month, int *day);
 extern int	date2j(int year, int month, int day);
 
@@ -299,31 +285,28 @@ extern int ParseDateTime(const char *timestr, char *workbuf, size_t buflen,
 			  int maxfields, int *numfields);
 extern int DecodeDateTime(char **field, int *ftype,
 			   int nf, int *dtype,
-			   struct tm * tm, fsec_t *fsec, int *tzp);
+			   struct pg_tm * tm, fsec_t *fsec, int *tzp);
 extern int DecodeTimeOnly(char **field, int *ftype,
 			   int nf, int *dtype,
-			   struct tm * tm, fsec_t *fsec, int *tzp);
+			   struct pg_tm * tm, fsec_t *fsec, int *tzp);
 extern int DecodeInterval(char **field, int *ftype,
 			   int nf, int *dtype,
-			   struct tm * tm, fsec_t *fsec);
+			   struct pg_tm * tm, fsec_t *fsec);
 extern void DateTimeParseError(int dterr, const char *str,
-							   const char *datatype);
+				   const char *datatype);
 
-extern int	DetermineLocalTimeZone(struct tm * tm);
+extern int	DetermineTimeZoneOffset(struct pg_tm * tm, pg_tz *tzp);
 
-extern int	EncodeDateOnly(struct tm * tm, int style, char *str);
-extern int	EncodeTimeOnly(struct tm * tm, fsec_t fsec, int *tzp, int style, char *str);
-extern int	EncodeDateTime(struct tm * tm, fsec_t fsec, int *tzp, char **tzn, int style, char *str);
-extern int	EncodeInterval(struct tm * tm, fsec_t fsec, int style, char *str);
+extern int	EncodeDateOnly(struct pg_tm * tm, int style, char *str);
+extern int	EncodeTimeOnly(struct pg_tm * tm, fsec_t fsec, int *tzp, int style, char *str);
+extern int	EncodeDateTime(struct pg_tm * tm, fsec_t fsec, int *tzp, char **tzn, int style, char *str);
+extern int	EncodeInterval(struct pg_tm * tm, fsec_t fsec, int style, char *str);
 
 extern int	DecodeSpecial(int field, char *lowtoken, int *val);
 extern int	DecodeUnits(int field, char *lowtoken, int *val);
-extern bool ClearDateCache(bool newval, bool doit, bool interactive);
 
 extern int	j2day(int jd);
 
-extern int	DecodePosixTimezone(char *str, int *tzp);
-
 extern bool CheckDateTokenTables(void);
 
 #endif   /* DATETIME_H */
